# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from multipledispatch import dispatch
from collections import Iterable
from .base import BaseObject, BaseMatrix, BaseMatrix33, BaseMatrix44, BaseQuaternion, BaseVector, NpProxy
from .. import matrix44

class Matrix44(BaseMatrix44):
    _module = matrix44
    _shape = (4,4,)

    # m<c> style access
    m1 = NpProxy(0)
    m2 = NpProxy(1)
    m3 = NpProxy(2)
    m4 = NpProxy(3)

    # m<r><c> access
    m11 = NpProxy((0,0))
    m12 = NpProxy((0,1))
    m13 = NpProxy((0,2))
    m14 = NpProxy((0,3))
    m21 = NpProxy((1,0))
    m22 = NpProxy((1,1))
    m23 = NpProxy((1,2))
    m24 = NpProxy((1,3))
    m31 = NpProxy((2,0))
    m32 = NpProxy((2,1))
    m33 = NpProxy((2,2))
    m34 = NpProxy((2,3))
    m41 = NpProxy((3,0))
    m42 = NpProxy((3,1))
    m43 = NpProxy((3,2))
    m44 = NpProxy((3,3))

    # rows
    r1 = NpProxy(0)
    r2 = NpProxy(1)
    r3 = NpProxy(2)
    r4 = NpProxy(3)

    # columns
    c1 = NpProxy((slice(0,4),0))
    c2 = NpProxy((slice(0,4),1))
    c3 = NpProxy((slice(0,4),2))
    c4 = NpProxy((slice(0,4),3))

    ########################
    # Creation
    @classmethod
    def from_matrix33(cls, matrix, dtype=None):
        return cls(matrix44.create_from_matrix33(matrix, dtype))

    @classmethod
    def perspective_projection(cls, fovy, aspect, near, far, dtype=None):
        return cls(matrix44.create_perspective_projection_matrix(fovy, aspect, near, far, dtype))

    @classmethod
    def perspective_projection_bounds(cls, left, right, top, bottom, near, far, dtype=None):
        return cls(matrix44.create_perspective_projection_matrix_from_bounds(left, right, top, bottom, near, far, dtype))

    @classmethod
    def orthogonal_projection(cls, left, right, top, bottom, near, far, dtype=None):
        return cls(matrix44.create_orthogonal_projection_matrix(left, right, top, bottom, near, far, dtype))

    @classmethod
    def from_translation(cls, translation, dtype=None):
        return cls(matrix44.create_from_translation(translation, dtype=dtype))

    @classmethod
    def from_quaternion(cls, quat, dtype=None):
        return cls(matrix44.create_from_quaternion(quat, dtype=dtype))

    def __new__(cls, value=None, dtype=None):
        if value is not None:
            obj = value
            if not isinstance(value, np.ndarray):
                obj = np.array(value, dtype=dtype)

            # matrix33
            if obj.shape == (3,3) or isinstance(obj, Matrix33):
                obj = matrix44.create_from_matrix33(obj, dtype=dtype)
            # quaternion
            elif obj.shape == (4,) or isinstance(obj, Quaternion):
                obj = matrix44.create_from_quaternion(obj, dtype=dtype)
        else:
            obj = np.zeros(cls._shape, dtype=dtype)
        obj = obj.view(cls)
        return super(Matrix44, cls).__new__(cls, obj)

    ########################
    # Basic Operators
    @dispatch(BaseObject)
    def __add__(self, other):
        raise ValueError('Cannot {} a {} to a {}'.format('add', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __sub__(self, other):
        raise ValueError('Cannot {} a {} from a {}'.format('subtract', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __mul__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('multiply', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __truediv__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __div__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    def __invert__(self):
        return self.inverse

    ########################
    # Matrices
    @dispatch(BaseMatrix)
    def __mul__(self, other):
        return Matrix44(matrix44.multiply(self, other.matrix44))

    ########################
    # Quaternions
    @dispatch(BaseQuaternion)
    def __mul__(self, other):
        m = other.matrix44
        return self * m

    ########################
    # Vectors
    @dispatch(BaseVector)
    def __mul__(self, other):
        return other.__class__(matrix44.apply_to_vector(self, other))

    ########################
    # Methods and Properties
    @property
    def matrix33(self):
        return Matrix33(self)

    @property
    def matrix44(self):
        return self

    @property
    def quaternion(self):
        return Quaternion(self)

    @property
    def inverse(self):
        return Matrix44(matrix44.inverse(self))

    @property
    def translation(self):
        return Vector4(self[3,0:4])

from .matrix33 import Matrix33
from .quaternion import Quaternion



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__all__ = [
    'base',
    'matrix33',
    'matrix44',
    'quaternion',
    'vector3',
    'vector4',
]

from . import (
    base,
    matrix33,
    matrix44,
    quaternion,
    vector3,
    vector4,
)

from .matrix33 import Matrix33
from .matrix44 import Matrix44
from .quaternion import Quaternion
from .vector3 import Vector3
from .vector4 import Vector4




# -*- coding: utf-8 -*-
"""Geometry functions.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


def create_quad(scale=(1.0,1.0), st=False, rgba=False, dtype='float32', type='triangles'):
    """Returns a Quad reading for rendering.

    Output is a tuple of numpy arrays.
    The first value is the vertex data, the second is the indices.

    The first dimension of the vertex data contains the list of vertices.
    The second dimension is the vertex data.

    Vertex data is always in the following order::

        [x, y, z, s, t, r, g, b, a]

    ST and RGBA are optional.
    If ST is dropped but RGBA is included the format will be::

        [x, y, z, r, g, b, a]

    If both ST and RGBA are dropped the format will be::

        [x, y, z]

    RGBA can also be of size 3 (RGB) or 4 (RGBA).

    Output format is as follows::

        numpy.array([
            # vertex 1
            [x, y, z, s, t, r, g, b, a],
            # vertex 2
            [x, y, z, s, t, r, g, b, a],
            ...
            # vertex N
            [x, y, z, s, t, r, g, b, a],
        ], dtype = dtype)


    :param bool,scalar,list,tuple,numpy.ndarray st: The ST texture co-ordinates.

        Default is False, which means ST will not be included in the array.

        If True is passed, the default ST values will be provided with the bottom-left
        of the quad being located at ST=(0.0,0.0) and the top-right being located at
        ST=(1.0,1.0).

        If a 2d list, tuple or numpy array is passed, it must have one of the following
        shapes::

                (2,2,), (4,2,),

        If the shape is (2,2,), the values are interpreted as the minimum and maximum
        values for ST.

        For example::

            st=((0.1,0.3),(0.2,0.4))

        S values will be between 0.1 to 0.2.
        T values will be between 0.3 to 0.4.

        The bottom left will receive the minimum of both, and the top right will receive
        the maximum.

        If the shape is (4,2,), the values are interpreted as being the actual ST values
        for the 4 vertices of the Quad.

        The vertices are in counter-clockwise winding order from the top right::

            [top-right, top-left, bottom-left, bottom-right,]

    :param bool,scalar,list,tuple,numpy.ndarray rgba: The RGBA colour.

        Default is False, which means RGBA will not be included in the array.

        If True is passed, the default RGBA values will be provided with all vertices
        being RGBA=(1.0, 1.0, 1.0, 1.0)

        If a 2d list, tuple or numpy array is passed, it must have one of the following
        shapes::

            (3,), (4,), (4,3,), (4,4,),

        If the shape is (3,), the values are interpreted as being an RGB value (no alpha)
        to set on all vertices.

        If the shape is (4,), the values are intepreted the same as the shape (3,) except
        the alpha value is included.

        If the shape is (4,3,), the values are interpreted as being a colour to set on
        the 4 vertices of the Quad.

        The vertices are in counter-clockwise winding order from the top right::

            [top-right, top-left, bottom-left, bottom-right]

        If the shape is (4,4,), the values are intepreted the same as the shape (4,3,)
        except the alpha value is included.

    :param string type: The type of indices to generate.

        Valid values are::

            ['triangles', 'triangle_strip', 'triangle_fan', 'quads', 'quad_strip',]

        If you just want the vertices without any index manipulation, use 'quads'.

    """
    shape = [4, 3]
    rgba_offset = 3

    width, height = scale
    # half the dimensions
    width /= 2.0
    height /= 2.0

    vertices = np.array([
        # top right
        ( width, height, 0.0,),
        # top left
        (-width, height, 0.0,),
        # bottom left
        (-width,-height, 0.0,),
        # bottom right
        ( width,-height, 0.0,),
    ], dtype=dtype)

    st_values = None
    rgba_values = None

    if st:
        # default st values
        st_values = np.array([
            (1.0, 1.0,),
            (0.0, 1.0,),
            (0.0, 0.0,),
            (1.0, 0.0,),
        ], dtype=dtype)

        if isinstance(st, bool):
            pass
        elif isinstance(st, (int, float)):
            st_values *= st
        elif isinstance(st, (list, tuple, np.ndarray)):
            st = np.array(st, dtype=dtype)
            if st.shape == (2,2,):
                # min / max
                st_values *= st[1] - st[0]
                st_values += st[0]
            elif st.shape == (4,2,):
                # st values specified manually
                st_values[:] = st
            else:
                raise ValueError('Invalid shape for st')
        else:
            raise ValueError('Invalid value for st')

        shape[-1] += st_values.shape[-1]
        rgba_offset += st_values.shape[-1]

    if rgba:
        # default rgba values
        rgba_values = np.tile(np.array([1.0, 1.0, 1.0, 1.0], dtype=dtype), (4,1,))

        if isinstance(rgba, bool):
            pass
        elif isinstance(rgba, (int, float)):
            # int / float expands to RGBA with all values == value
            rgba_values *= rgba 
        elif isinstance(rgba, (list, tuple, np.ndarray)):
            rgba = np.array(rgba, dtype=dtype)

            if rgba.shape == (3,):
                rgba_values = np.tile(rgba, (4,1,))
            elif rgba.shape == (4,):
                rgba_values[:] = rgba
            elif rgba.shape == (4,3,):
                rgba_values = rgba
            elif rgba.shape == (4,4,):
                rgba_values = rgba
            else:
                raise ValueError('Invalid shape for rgba')
        else:
            raise ValueError('Invalid value for rgba')

        shape[-1] += rgba_values.shape[-1]

    data = np.empty(shape, dtype=dtype)
    data[:,:3] = vertices
    if st_values is not None:
        data[:,3:5] = st_values
    if rgba_values is not None:
        data[:,rgba_offset:] = rgba_values

    if type == 'triangles':
        # counter clockwise
        # top right -> top left -> bottom left
        # top right -> bottom left -> bottom right
        indices = np.array([0, 1, 2, 0, 2, 3])
    elif type == 'triangle_strip':
        # verify
        indices = np.arange(len(data))
    elif type == 'triangle_fan':
        # verify
        indices = np.arange(len(data))
    elif type == 'quads':
        indices = np.arange(len(data))
    elif type == 'quad_strip':
        indices = np.arange(len(data))
    else:
        raise ValueError('Unknown type')

    return data, indices

def create_cube(scale=(1.0,1.0,1.0), st=False, rgba=False, dtype='float32', type='triangles'):
    """Returns a Cube reading for rendering.

    Output is a tuple of numpy arrays.
    The first value is the vertex data, the second is the indices.

    The first dimension of the vertex data contains the list of vertices.
    The second dimension is the vertex data.

    Vertex data is always in the following order::

        [x, y, z, s, t, r, g, b, a]

    ST and RGBA are optional.
    If ST is dropped but RGBA is included the format will be::

        [x, y, z, r, g, b, a]

    If both ST and RGBA are dropped the format will be::

        [x, y, z]

    RGBA can also be of size 3 (RGB) or 4 (RGBA).

    Output format is as follows::

        numpy.array([
            # vertex 1
            [x, y, z, s, t, r, g, b, a],
            # vertex 2
            [x, y, z, s, t, r, g, b, a],
            ...
            # vertex N
            [x, y, z, s, t, r, g, b, a],
        ], dtype = dtype)

    :param bool,scalar,list,tuple,numpy.ndarray st: The ST texture co-ordinates.

        Default is False, which means ST will not be included in the array.

        If True is passed, the default ST values will be provided with the bottom-left
        of the quad being located at ST=(0.0,0.0) and the top-right being located at
        ST=(1.0,1.0).

        If a 2d list, tuple or numpy array is passed, it must have one of the following
        shapes::

            (2,2,), (4,2,), (6,2,),

        If the shape is (2,2,), the values are interpreted as the minimum and maximum
        values for ST.

        For example::

            st=((0.1,0.3),(0.2,0.4))

        S values will be between 0.1 to 0.2.
        T values will be between 0.3 to 0.4.

        The bottom left will receive the minimum of both, and the top right will receive
        the maximum.

        If the shape is (4,2,), the values are interpreted as being the actual ST values
        for the 4 vertices of each face.

        The vertices are in counter-clockwise winding order from the top right::

            [top-right, top-left, bottom-left, bottom-right,]

        If the shape is (6,2,), the values are interpreted as being the minimum and maximum
        values for each face of the cube.

        The faces are in the following order::

            [front, right, back, left, top, bottom,]


    :param bool,scalar,list,tuple,numpy.ndarray rgba: The RGBA colour.

        Default is False, which means RGBA will not be included in the array.

        If True is passed, the default RGBA values will be provided with all vertices
        being RGBA=(1.0, 1.0, 1.0, 1.0).

        If a 2d list, tuple or numpy array is passed, it must have one of the following
        shapes.::

                (3,), (4,), (4,3,), (4,4,), (6,3,), (6,4,), (24,3,), (24,4,),

        If the shape is (3,), the values are interpreted as being an RGB value (no alpha)
        to set on all vertices.

        If the shape is (4,), the values are intepreted the same as the shape (3,) except
        the alpha value is included.

        If the shape is (4,3,), the values are interpreted as being a colour to set on
        the 4 vertices of each face.

        The vertices are in counter-clockwise winding order from the top right::

            [top-right, top-left, bottom-left, bottom-right]

        If the shape is (4,4,), the values are intepreted the same as the shape (4,3,)
        except the alpha value is included.

        If the shape is (6,3,), the values are interpreted as being one RGB value (no alpha)
        for each face.

        The faces are in the following order::

            [front, right, back, left, top, bottom,]

        If the shape is (6,4,), the values are interpreted the same as the shape (6,3,) except
        the alpha value is included.

        If the shape is (24,3,), the values are interpreted as being an RGB value (no alpha)
        to set on each vertex of each face (4 * 6).

        The faces are in the following order::

            [front, right, back, left, top, bottom,]

        The vertices are in counter-clockwise winding order from the top right::

            [top-right, top-left, bottom-left, bottom-right]

        If the shape is (24,4,), the values are interpreted the same as the shape (24,3,)
        except the alpha value is included.


    :param string type: The type of indices to generate.

        Valid values are::

            ['triangles', 'triangle_strip', 'triangle_fan', 'quads', 'quad_strip',]

        If you just want the vertices without any index manipulation, use 'quads'.
    """

    shape = [24, 3]
    rgba_offset = 3

    width, height, depth = scale
    # half the dimensions
    width /= 2.0
    height /= 2.0
    depth /= 2.0

    vertices = np.array([
        # front
        # top right
        ( width, height, depth,),
        # top left
        (-width, height, depth,),
        # bottom left
        (-width,-height, depth,),
        # bottom right
        ( width,-height, depth,),

        # right
        # top right
        ( width, height,-depth),
        # top left
        ( width, height, depth),
        # bottom left
        ( width,-height, depth),
        # bottom right
        ( width,-height,-depth),

        # back
        # top right
        (-width, height,-depth),
        # top left
        ( width, height,-depth),
        # bottom left
        ( width,-height,-depth),
        # bottom right
        (-width,-height,-depth),

        # left
        # top right
        (-width, height, depth),
        # top left
        (-width, height,-depth),
        # bottom left
        (-width,-height,-depth),
        # bottom right
        (-width,-height, depth),

        # top
        # top right
        ( width, height,-depth),
        # top left
        (-width, height,-depth),
        # bottom left
        (-width, height, depth),
        # bottom right
        ( width, height, depth),

        # bottom
        # top right
        ( width,-height, depth),
        # top left
        (-width,-height, depth),
        # bottom left
        (-width,-height,-depth),
        # bottom right
        ( width,-height,-depth),
    ], dtype=dtype)

    st_values = None
    rgba_values = None

    if st:
        # default st values
        st_values = np.tile(
            np.array([
                (1.0, 1.0,),
                (0.0, 1.0,),
                (0.0, 0.0,),
                (1.0, 0.0,),
            ], dtype=dtype),
            (6,1,)
        )

        if isinstance(st, bool):
            pass
        elif isinstance(st, (int, float)):
            st_values *= st
        elif isinstance(st, (list, tuple, np.ndarray)):
            st = np.array(st, dtype=dtype)
            if st.shape == (2,2,):
                # min / max
                st_values *= st[1] - st[0]
                st_values += st[0]
            elif st.shape == (4,2,):
                # per face st values specified manually
                st_values[:] = np.tile(st, (6,1,))
            elif st.shape == (6,2,):
                # st values specified manually
                st_values[:] = st
            else:
                raise ValueError('Invalid shape for st')
        else:
            raise ValueError('Invalid value for st')

        shape[-1] += st_values.shape[-1]
        rgba_offset += st_values.shape[-1]

    if rgba:
        # default rgba values
        rgba_values = np.tile(np.array([1.0, 1.0, 1.0, 1.0], dtype=dtype), (24,1,))

        if isinstance(rgba, bool):
            pass
        elif isinstance(rgba, (int, float)):
            # int / float expands to RGBA with all values == value
            rgba_values *= rgba 
        elif isinstance(rgba, (list, tuple, np.ndarray)):
            rgba = np.array(rgba, dtype=dtype)

            if rgba.shape == (3,):
                rgba_values = np.tile(rgba, (24,1,))
            elif rgba.shape == (4,):
                rgba_values[:] = np.tile(rgba, (24,1,))
            elif rgba.shape == (4,3,):
                rgba_values = np.tile(rgba, (6,1,))
            elif rgba.shape == (4,4,):
                rgba_values = np.tile(rgba, (6,1,))
            elif rgba.shape == (6,3,):
                rgba_values = np.repeat(rgba, 4, axis=0)
            elif rgba.shape == (6,4,):
                rgba_values = np.repeat(rgba, 4, axis=0)
            elif rgba.shape == (24,3,):
                rgba_values = rgba
            elif rgba.shape == (24,4,):
                rgba_values = rgba
            else:
                raise ValueError('Invalid shape for rgba')
        else:
            raise ValueError('Invalid value for rgba')

        shape[-1] += rgba_values.shape[-1]

    data = np.empty(shape, dtype=dtype)
    data[:,:3] = vertices
    if st_values is not None:
        data[:,3:5] = st_values
    if rgba_values is not None:
        data[:,rgba_offset:] = rgba_values

    if type == 'triangles':
        # counter clockwise
        # top right -> top left -> bottom left
        # top right -> bottom left -> bottom right
        indices = np.tile(np.array([0, 1, 2, 0, 2, 3], dtype='int'), (6,1))
        for face in range(6):
            indices[face] += (face * 4)
        indices.shape = (-1,)
    elif type == 'triangle_strip':
        raise NotImplementedError
    elif type == 'triangle_fan':
        raise NotImplementedError
    elif type == 'quads':
        raise NotImplementedError
    elif type == 'quad_strip':
        raise NotImplementedError
    else:
        raise ValueError('Unknown type')

    return data, indices



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from multipledispatch import dispatch
from collections import Iterable
from .base import BaseObject, BaseMatrix, BaseMatrix33, BaseMatrix44, BaseQuaternion, BaseVector, NpProxy
from .. import matrix33

class Matrix33(BaseMatrix33):
    _module = matrix33
    _shape = (3,3,)

    # m<c> style access
    m1 = NpProxy(0)
    m2 = NpProxy(1)
    m3 = NpProxy(2)

    # m<r><c> access
    m11 = NpProxy((0,0))
    m12 = NpProxy((0,1))
    m13 = NpProxy((0,2))
    m21 = NpProxy((1,0))
    m22 = NpProxy((1,1))
    m23 = NpProxy((1,2))
    m31 = NpProxy((2,0))
    m32 = NpProxy((2,1))
    m33 = NpProxy((2,2))

    # rows
    r1 = NpProxy(0)
    r2 = NpProxy(1)
    r3 = NpProxy(2)

    # columns
    c1 = NpProxy((slice(0,3),0))
    c2 = NpProxy((slice(0,3),1))
    c3 = NpProxy((slice(0,3),2))

    ########################
    # Creation
    @classmethod
    def from_matrix44(cls, matrix, dtype=None):
        return cls(matrix33.create_from_matrix44(matrix, dtype))

    def __new__(cls, value=None, dtype=None):
        if value is not None:
            obj = value
            if not isinstance(value, np.ndarray):
                obj = np.array(value, dtype=dtype)

            # matrix44
            if obj.shape == (4,4) or isinstance(obj, Matrix44):
                obj = matrix33.create_from_matrix44(obj, dtype=dtype)
            # quaternion
            elif obj.shape == (4,) or isinstance(obj, Quaternion):
                obj = matrix33.create_from_quaternion(obj, dtype=dtype)
        else:
            obj = np.zeros(cls._shape, dtype=dtype)
        obj = obj.view(cls)
        return super(Matrix33, cls).__new__(cls, obj)

    ########################
    # Basic Operators
    @dispatch(BaseObject)
    def __add__(self, other):
        raise ValueError('Cannot {} a {} to a {}'.format('add', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __sub__(self, other):
        raise ValueError('Cannot {} a {} from a {}'.format('subtract', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __mul__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('multiply', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __truediv__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __div__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    def __invert__(self):
        return self.inverse

    ########################
    # Matrices
    @dispatch(BaseMatrix)
    def __mul__(self, other):
        return Matrix33(matrix33.multiply(self, other.matrix33))

    ########################
    # Quaternions
    @dispatch(BaseQuaternion)
    def __mul__(self, other):
        m = other.matrix33
        return self * m
 
    ########################
    # Vectors
    @dispatch(BaseVector)
    def __mul__(self, other):
        return other.__class__(matrix33.apply_to_vector(self, other))

    ########################
    # Methods and Properties
    @property
    def matrix33(self):
        return self

    @property
    def matrix44(self):
        return Matrix44(self)

    @property
    def quaternion(self):
        return Quaternion(self)

    @property
    def inverse(self):
        return Matrix33(matrix33.inverse(self))

from .matrix44 import Matrix44
from .quaternion import Quaternion



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# the version of software
# this is used by the setup.py script
from .version import __version__

__all__ = [
    'aabb',
    'aambb',
    'euler',
    'geometric_tests',
    'geometry',
    'integer',
    'line',
    'matrix33',
    'matrix44',
    'plane',
    'quaternion',
    'ray',
    'rectangle',
    'sphere',
    'trig',
    'utils',
    'vector',
    'vector3',
    'vector4',
    'Matrix33',
    'Matrix44',
    'Quaternion',
    'Vector3',
    'Vector4',
]

from . import (
    aabb,
    aambb,
    euler,
    geometric_tests,
    geometry,
    integer,
    line,
    matrix33,
    matrix44,
    plane,
    quaternion,
    ray,
    rectangle,
    sphere,
    trig,
    utils,
    vector,
    vector3,
    vector4,
)

from .objects import (
    Matrix33,
    Matrix44,
    Quaternion,
    Vector3,
    Vector4
)

# because of circular imports, we cannot put these inside each module
# so insert them here
setattr(matrix33, 'Matrix33', Matrix33)
setattr(matrix44, 'Matrix44', Matrix44)
setattr(quaternion, 'Quaternion', Quaternion)
setattr(vector3, 'Vector3', Vector3)
setattr(vector4, 'Vector4', Vector4)




# -*- coding: utf-8 -*-
"""Defines a number of functions to test interactions between
various forms data types.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import math
import numpy as np
from . import rectangle, vector, vector3
from .utils import all_parameters_as_numpy_arrays, parameters_as_numpy_arrays

"""
TODO: line_intersect_plane
TODO: line_segment_intersect_plane
TODO: ray_intersect_ray
TODO: line_intersect_line
TODO: line_segment_intersect_line_segment
"""

@all_parameters_as_numpy_arrays
def point_intersect_line(point, line):
    """Calculates the intersection point of a point and aline.

    Performed by checking if the cross-product
    of the point relative to the line is
    0.
    """
    rl = line[1] - line[0]
    rp = point - line[0]
    cross = vector3.cross(rl, rp)

    # check if the cross product is zero
    if np.count_nonzero(cross) > 0:
        return None
    return point

@all_parameters_as_numpy_arrays
def point_intersect_line_segment(point, line):
    """Calculates the intersection point of a point and a line segment.

    Performed by checking if the cross-product
    of the point relative to the line is
    0 and if the dot product of the point
    relative to the line start AND the end
    point relative to the line start is
    less than the segment's squared length.
    """
    rl = line[1] - line[0]
    rp = point - line[0]
    cross = vector3.cross(rl, rp)
    dot = vector.dot(rp, rl)
    squared_length = vector.squared_length(rl)

    if np.count_nonzero(cross) > 0:
        return None

    if dot < 0.0 or dot > squared_length:
        return None
    return point

@all_parameters_as_numpy_arrays
def point_intersect_rectangle(point, rect):
    """Calculates the intersection point of a point and a 2D rectangle.

    For 3D points, the Z axis will be ignored.

    :return: Returns True if the point is touching
    or within the rectangle.
    """
    left, right, bottom, top = rectangle.bounds(rect)
    if \
        point[0] < left or \
        point[0] > right or \
        point[1] < bottom or \
        point[1] > top:
        return None
    return point

@parameters_as_numpy_arrays('ray', 'plane')
def ray_intersect_plane(ray, plane, front_only=False):
    """Calculates the intersection point of a ray and a plane.

    :param numpy.array ray: The ray to test for intersection.
    :param numpy.array plane: The ray to test for intersection.
    :param boolean front_only: Specifies if the ray should
    only hit the front of the plane.
    Collisions from the rear of the plane will be
    ignored.

    :return The intersection point, or None
    if the ray is parallel to the plane.
    Returns None if the ray intersects the back
    of the plane and front_only is True.
    """
    """
    Distance to plane is defined as
    t = (pd - p0.n) / rd.n
    where:
    rd is the ray direction
    pd is the point on plane . plane normal
    p0 is the ray position
    n is the plane normal

    if rd.n == 0, the ray is parallel to the
    plane.
    """
    p = plane[:3] * plane[3]
    n = plane[:3]
    rd_n = vector.dot(ray[1], n)

    if rd_n == 0.0:
        return None

    if front_only == True:
        if rd_n >= 0.0:
            return None

    pd = vector.dot(p, n)
    p0_n = vector.dot(ray[0], n)
    t = (pd - p0_n) / rd_n
    return ray[0] + (ray[1] * t)

@all_parameters_as_numpy_arrays
def point_closest_point_on_ray(point, ray):
    """Calculates the point on a ray that is closest to a point.

    :param numpy.array point: The point to check with.
    :param numpy.array ray: The ray to check against.
    :rtype: numpy.array
    :return: The closest point on the ray to the point.
    """
    """
    t = (p - rp).n
    cp = rp + (n * t)
    where
    p is the point
    rp is the ray origin
    n is the ray normal of unit length
    t is the distance along the ray to the point
    """
    normalised_n = vector.normalise(ray[1])
    relative_point = (point - ray[0])
    t = vector.dot(relative_point, normalised_n)
    return ray[0] + (normalised_n * t)

@all_parameters_as_numpy_arrays
def point_closest_point_on_line(point, line):
    """Calculates the point on the line that is closest to
    the specified point.

    :param numpy.array point: The point to check with.
    :param numpy.array line: The line to check against.
    :rtype: numpy.array
    :return: The closest point on the line to the point.
    """
    """
    rl = va->b (relative line)
    rp = va->p (relative point)
    u' = u / |u| (normalise)
    cp = a + (u' * (u'.v))
    where:
    a = line start
    b = line end
    p = point
    cp = closest point
    """
    rl = line[1] - line[0]
    rp = point - line[0]
    rl = vector.normalise(rl)
    dot = vector.dot(rl, rp)
    return line[0] + (rl * dot)

@all_parameters_as_numpy_arrays
def point_closest_point_on_line_segment(point, segment):
    """Calculates the point on the line segment that is closest
    to the specified point.

    This is similar to point_closest_point_on_line, except this
    is against the line segment of finite length. Whereas point_closest_point_on_line
    checks against a line of infinite length.

    :param numpy.array point: The point to check with.
    :param numpy.array line_segment: The finite line segment to check against.
    :rtype: numpy.array
    :return: The closest point on the line segment to the point.
    """
    # check if the line has any length
    rl = segment[1] - segment[0]
    squared_length = vector.squared_length(rl)
    if squared_length == 0.0:
        return segment[0]

    rp = point - segment[0]
    # check that / squared_length is correct
    dot = vector.dot(rp, rl) / squared_length;

    if dot < 0.0:
        return segment[0]
    elif dot > 1.0:
        return segment[1]

    # within segment
    # perform the same calculation as closest_point_on_line
    return segment[0] + (rl * dot)

@all_parameters_as_numpy_arrays
def vector_parallel_vector(v1, v2):
    """Checks if two vectors are parallel.

    :param numpy.array v1, v2: The vectors to check.
    :rtype: boolean
    :return: Returns True if the two vectors are parallel.
    """
    # we cross product the 2 vectors
    # if the result is 0, then they are parallel
    cross = vector3.cross(v1, v2)
    return 0 == np.count_nonzero(cross)
    
@all_parameters_as_numpy_arrays
def ray_parallel_ray(ray1, ray2):
    """Checks if two rays are parallel.

    :param numpy.array ray1, ray2: The rays to check.
    :rtype: boolean
    :return: Returns True if the two rays are parallel.
    """
    # we use a cross product in-case the ray direction
    # isn't unit length
    return vector_parallel_vector(ray1[ 1 ], ray2[ 1 ])

@all_parameters_as_numpy_arrays
def ray_coincident_ray(ray1, ray2):
    """Check if rays are coincident.

    Rays must not only be parallel to each other, but reside
    along the same vector.

    :param numpy.array ray1, ray2: The rays to check.
    :rtype: boolean
    :return: Returns True if the two rays are co-incident.
    """
    # ensure the ray's directions are the same
    if ray_parallel_ray(ray1, ray2):
        # get the delta between the two ray's start point
        delta = ray2[0] - ray1[0]

        # get the cross product of the ray delta and
        # the direction of the rays
        cross = vector3.cross(delta, ray2[1])

        # if the cross product is zero, the start of the
        # second ray is in line with the direction of the
        # first ray
        if np.count_nonzero(cross) > 0:
            return False

        return True
    return False

@all_parameters_as_numpy_arrays
def ray_intersect_aabb(ray, aabb):
    """Calculates the intersection point of a ray and an AABB

    :param numpy.array ray1: The ray to check.
    :param numpy.array aabb: The Axis-Aligned Bounding Box to check against.
    :rtype: numpy.array
    :return: Returns a vector if an intersection occurs.
        Returns None if no intersection occurs.
    """
    """
    http://gamedev.stackexchange.com/questions/18436/most-efficient-aabb-vs-ray-collision-algorithms
    """
    # this is basically "numpy.divide( 1.0, ray[ 1 ] )"
    # except we're trying to avoid a divide by zero warning
    # so where the ray direction value is 0.0, just use infinity
    # which is what we want anyway
    direction = ray[1]
    dir_fraction = np.empty(3, dtype = ray.dtype)
    dir_fraction[direction == 0.0] = np.inf
    dir_fraction[direction != 0.0] = np.divide(1.0, direction[direction != 0.0])

    t1 = (aabb[0,0] - ray[0,0]) * dir_fraction[ 0 ]
    t2 = (aabb[1,0] - ray[0,0]) * dir_fraction[ 0 ]
    t3 = (aabb[0,1] - ray[0,1]) * dir_fraction[ 1 ]
    t4 = (aabb[1,1] - ray[0,1]) * dir_fraction[ 1 ]
    t5 = (aabb[0,2] - ray[0,2]) * dir_fraction[ 2 ]
    t6 = (aabb[1,2] - ray[0,2]) * dir_fraction[ 2 ]


    tmin = max(min(t1, t2), min(t3, t4), min(t5, t6))
    tmax = min(max(t1, t2), max(t3, t4), max(t5, t6))

    # if tmax < 0, ray (line) is intersecting AABB
    # but the whole AABB is behind the ray start
    if tmax < 0:
        return None

    # if tmin > tmax, ray doesn't intersect AABB
    if tmin > tmax:
        return None

    # t is the distance from the ray point
    # to intersection

    t = abs(tmin)
    point = ray[0] + (ray[1] * t)
    return point

@all_parameters_as_numpy_arrays
def point_height_above_plane(point, plane):
    """Calculates how high a point is above a plane.

    :param numpy.array point: The point to check.
    :param numpy.array plane: The plane to check.
    :rtype: float
    :return: The height above the plane as a float. The value will be
        negative if the point is behind the plane.
    """
    """
    http://www.vitutor.com/geometry/distance/point_plane.html
    d(P) = (AX + BY + CZ + D) / sqrt(A^2 + B^2 + C^2)

    Normal is unit length, so it's length is 1.0.
    Therefore, we can ignore the division all together.
    Just perform Pn . [XYZ1]
    """
    return np.dot(plane, [point[0], point[1], point[2], 1.0])

@all_parameters_as_numpy_arrays
def point_closest_point_on_plane(point, plane):
    """Calculates the point on a plane that is closest to a point.

    :param numpy.array point: The point to check with.
    :param numpy.array plane: The infinite plane to check against.
    :rtype: numpy.array
    :return: The closest point on the plane to the point.
    """
    """
    point on plane is defined as:
    q' = q + (d - q.n)n
    where:
    q' is the point on the plane
    q is the point we are checking
    d is the value of normal dot position
    n is the plane normal
    """
    n = plane[:3]
    p = n * plane[3]
    d = np.dot(p, n)
    qn = np.dot(point, n)
    return point + (n * (d - qn))

@all_parameters_as_numpy_arrays
def sphere_does_intersect_sphere(s1, s2):
    """Checks if two spheres overlap.

    Note: This will return True if the two spheres are
    touching perfectly but sphere_penetration_sphere
    will return 0.0 as the touch but don't penetrate.

    This is faster than circle_penetrate_amount_circle
    as it avoids a square root calculation.

    :param numpy.array s1: The first circle.
    :param numpy.array s2: The second circle.
    :rtype: boolean
    :return: Returns True if the circles overlap.
        Otherwise, returns False.
    """
    delta = s2[:3] - s1[:3]
    distance_squared = vector.squared_length(delta)

    radii_squared = math.pow(s1[3] + s2[3], 2.0)

    if distance_squared > radii_squared:
        return False
    return True

@all_parameters_as_numpy_arrays
def sphere_penetration_sphere(s1, s2):
    """Calculates the distance two spheres have penetrated
    into one another.

    :param numpy.array s1: The first circle.
    :param numpy.array s2: The second circle.
    :rtype: float
    :return: The total overlap of the two spheres.
        This is essentially:
        r1 + r2 - distance 
        Where r1 and r2 are the radii of circle 1 and 2
        and distance is the length of the vector p2 - p1.
        Will return 0.0 if the circles do not overlap.
    """
    delta = s2[:3] - s1[:3]
    distance = vector.length(delta)

    combined_radii = s1[3] + s2[3]
    penetration = combined_radii - distance

    if penetration <= 0.0:
        return 0.0
    return penetration



# -*- coding: utf-8 -*-
"""Provide functions for the manipulation of integers.
"""

def count_bits(value):
    """Counts the number of bits set to 1 in an integer.

    For example::

        >>> count_bits(0b101111)
        5
        >>> count_bits(0xf)
        4
        >>> count_bits(8)
        1
        >>> count_bits(3)
        2
    
    :param int value: An integer.
    :rtype: integer
    :return: The count of bits set to 1.

    .. seealso:: http://wiki.python.org/moin/BitManipulation
    """
    count = 0
    while (value):
        count += (value & 1)
        value >>= 1
    
    return count




# -*- coding: utf-8 -*-
"""Provide functions for the creation and manipulation of Planes.

Planes are represented using a numpy.array of shape (4,).
The values represent the plane equation using the values A,B,C,D.

The first three values are the normal vector.
The fourth value is the distance of the plane from the origin, down the normal.

.. seealso: http://en.wikipedia.org/wiki/Plane_(geometry)
.. seealso: http://mathworld.wolfram.com/Plane.html
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from . import vector
from .utils import all_parameters_as_numpy_arrays, parameters_as_numpy_arrays


def create(normal=None, distance=0.0, dtype=None):
    """Creates a plane that runs along the X,Y plane.

    It crosses the origin with a normal of 0,0,1 (+Z).

    :rtype: numpy.array
    :return: A plane that runs along the X,Y plane.
    """
    if normal is None:
        normal = [0.0, 0.0, 1.0]
    return np.array([normal[0], normal[1], normal[2], distance], dtype=dtype)

@parameters_as_numpy_arrays('vector1', 'vector2', 'vector3')
def create_from_points(vector1, vector2, vector3, dtype=None):
    """Create a plane from 3 co-planar vectors.

    The vectors must all lie on the same
    plane or an exception will be thrown.

    The vectors must not all be in a single line or
    the plane is undefined.

    The order the vertices are passed in will determine the
    normal of the plane.

    :param numpy.array vector1: a vector that lies on the desired plane.
    :param numpy.array vector2: a vector that lies on the desired plane.
    :param numpy.array vector3: a vector that lies on the desired plane.
    :raise ValueError: raised if the vectors are co-incident (in a single line).
    :rtype: numpy.array
    :return: A plane that contains the 3 specified vectors.
    """
    dtype = dtype or vector1.dtype

    # make the vectors relative to vector2
    relV1 = vector1 - vector2
    relV2 = vector3 - vector2
    
    # cross our relative vectors
    normal = np.cross(relV1, relV2)
    if np.count_nonzero(normal) == 0:
        raise ValueError("Vectors are co-incident")
    
    # create our plane
    return create_from_position(position=vector2, normal=normal, dtype=dtype)

@parameters_as_numpy_arrays('position', 'normal')
def create_from_position(position, normal, dtype=None):
    """Creates a plane at position with the normal being above the plane
    and up being the rotation of the plane.

    :param numpy.array position: The position of the plane.
    :param numpy.array normal: The normal of the plane. Will be normalised
        during construction.
    :rtype: numpy.array
    :return: A plane that crosses the specified position with the specified
        normal.
    """
    dtype = dtype or position.dtype
    # -d = a * px  + b * py + c * pz
    n = vector.normalise(normal)
    d = -np.sum(n * position)
    return create(n, d, dtype)

def create_xy(invert=False, distance=0., dtype=None):
    """Create a plane on the XY plane, starting at the origin with +Z being
    the up vector.

    The distance is the distance along the normal (-Z if inverted, otherwise +Z).
    """
    invert = -1. if invert else 1.
    return np.array([0., 0., 1. * invert, distance])

def create_xz(invert=False, distance=0., dtype=None):
    """Create a plane on the XZ plane, starting at the origin with +Y being
    the up vector.

    The distance is the distance along the normal (-Y if inverted, otherwise +Y).
    """
    invert = -1. if invert else 1.
    return np.array([0., 1. * invert, 0., distance])

def create_yz(invert=False, distance=0., dtype=None):
    """Create a plane on the YZ plane, starting at the origin with +X being
    the up vector.

    The distance is the distance along the normal (-X if inverted, otherwise +X).
    """
    invert = -1. if invert else 1.
    return np.array([1. * invert, 0., 0., distance])

def invert_normal(plane):
    """Flips the normal of the plane.

    The plane is **not** changed in place.

    :rtype: numpy.array
    :return: The plane with the normal inverted.
    """
    # flip the normal, and the distance
    return -plane

def position(plane):
    """Extracts the position vector from a plane.

    This will be a vector co-incident with the plane's normal.

    :param numpy.array plane: The plane.
    :rtype: numpy.array
    :return: A valid position that lies on the plane.
    """
    return plane[:3] * plane[3]

def normal(plane):
    """Extracts the normal vector from a plane.

    :param numpy.array plane: The plane.
    :rtype: numpy.array
    :return: The normal vector of the plane.
    """
    return plane[:3].copy()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from multipledispatch import dispatch
from .base import BaseObject, BaseQuaternion, BaseMatrix, BaseVector, NpProxy
from .. import quaternion

class Quaternion(BaseQuaternion):
    _module = quaternion
    _shape = (4,)

    x = NpProxy(0)
    y = NpProxy(1)
    z = NpProxy(2)
    w = NpProxy(3)
    xy = NpProxy([0,1])
    xyz = NpProxy([0,1,2])
    xyzw = NpProxy([0,1,2,3])
    xz = NpProxy([0,2])
    xzw = NpProxy([0,2,3])
    xyw = NpProxy([0,1,3])
    xw = NpProxy([0,3])

    ########################
    # Creation
    @classmethod
    def from_x_rotation(cls, theta, dtype=None):
        return cls(quaternion.create_from_x_rotation(theta, dtype))

    @classmethod
    def from_y_rotation(cls, theta, dtype=None):
        return cls(quaternion.create_from_y_rotation(theta, dtype))

    @classmethod
    def from_z_rotation(cls, theta, dtype=None):
        return cls(quaternion.create_from_z_rotation(theta, dtype))

    @classmethod
    def from_axis_rotation(cls, axis, theta, dtype=None):
        return cls(quaternion.create_from_axis_rotation(axis, theta, dtype))

    @classmethod
    def from_matrix(cls, matrix, dtype=None):
        return cls(quaternion.create_from_matrix(matrix, dtype))

    @classmethod
    def from_eulers(cls, eulers, dtype=None):
        return cls(quaternion.create_from_eulers(eulers, dtype))

    @classmethod
    def from_inverse_of_eulers(cls, eulers, dtype=None):
        return cls(quaternion.create_from_inverse_of_eulers(eulers, dtype))

    def __new__(cls, value=None, dtype=None):
        if value is not None:
            obj = value
            if not isinstance(value, np.ndarray):
                obj = np.array(value, dtype=dtype)

            # matrix33, matrix44
            if obj.shape in ((4,4,), (3,3,)) or isinstance(obj, (Matrix33, Matrix44)):
                obj = quaternion.create_from_matrix(obj, dtype=dtype)
        else:
            obj = quaternion.create(dtype=dtype)
        obj = obj.view(cls)
        return super(Quaternion, cls).__new__(cls, obj)

    ########################
    # Basic Operators
    @dispatch(BaseObject)
    def __add__(self, other):
        raise ValueError('Cannot {} a {} to a {}'.format('add', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __sub__(self, other):
        raise ValueError('Cannot {} a {} from a {}'.format('subtract', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __mul__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('multiply', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __truediv__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __div__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    ########################
    # Quaternions
    @dispatch(BaseQuaternion)
    def __mul__(self, other):
        return self.cross(other)

    @dispatch(BaseQuaternion)
    def __or__(self, other):
        return self.dot(other)

    def __invert__(self):
        return self.conjugate

    ########################
    # Matrices
    @dispatch(BaseMatrix)
    def __mul__(self, other):
        return self * Quaternion(other)

    ########################
    # Vectors
    @dispatch(BaseVector)
    def __mul__(self, other):
        return other.__class__(quaternion.apply_to_vector(self, other))

    ########################
    # Methods and Properties
    @property
    def length(self):
        return quaternion.length(self)

    def normalise(self):
        self[:] = quaternion.normalise(self)#self.normalised

    @property
    def normalised(self):
        return Quaternion(quaternion.normalise(self))

    @property
    def angle(self):
        return quaternion.rotation_angle(self)

    @property
    def axis(self):
        return Vector3(quaternion.rotation_axis(self))

    def cross(self, other):
        return Quaternion(quaternion.cross(self, other))

    def dot(self, other):
        return quaternion.dot(self, other)

    @property
    def conjugate(self):
        return Quaternion(quaternion.conjugate(self))

    @property
    def inverse(self):
        return Quaternion(quaternion.inverse(self))

    def power(self, exponent):
        return Quaternion(quaternion.power(self, exponent))

    @property
    def negative(self):
        return Quaternion(quaternion.negate(self))

    @property
    def is_identity(self):
        return quaternion.is_identity(self)

    @property
    def matrix44(self):
        return Matrix44.from_quaternion(self)

    @property
    def matrix33(self):
        return Matrix33.from_quaternion(self)

from .vector3 import Vector3
from .vector4 import Vector4
from .matrix33 import Matrix33
from .matrix44 import Matrix44



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from multipledispatch import dispatch
from collections import Iterable
from .base import BaseObject, BaseVector, BaseVector4, BaseQuaternion, BaseMatrix, NpProxy
from .. import vector4

# TODO: add < <= > >= == != operators

class Vector4(BaseVector4):
    _module = vector4
    _shape = (4,)

    x = NpProxy(0)
    y = NpProxy(1)
    z = NpProxy(2)
    w = NpProxy(3)
    xy = NpProxy([0,1])
    xyz = NpProxy([0,1,2])
    xyzw = NpProxy(slice(0,4))
    xz = NpProxy([0,2])
    xw = NpProxy([0,3])
    xzw = NpProxy([0,2,3])

    ########################
    # Creation
    def __new__(cls, value=None, dtype=None):
        if value is not None:
            obj = value
            if not isinstance(value, np.ndarray):
                obj = np.array(value, dtype=dtype)

            # vector3
            if obj.shape == (3,):
                obj = vector4.create_from_vector3(obj, dtype=dtype)
            # matrix44
            elif obj.shape == (4,4) or isinstance(obj, Matrix44):
                obj = vector4.create_from_matrix44_translation(obj, dtype=dtype)
        else:
            obj = np.zeros(cls._shape, dtype=dtype)
        obj = obj.view(cls)
        return super(Vector4, cls).__new__(cls, obj)

    ########################
    # Basic Operators
    @dispatch(BaseObject)
    def __add__(self, other):
        raise ValueError('Cannot {} a {} to a {}'.format('add', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __sub__(self, other):
        raise ValueError('Cannot {} a {} from a {}'.format('subtract', other.__class__.__name__, self.__class__.__name__))

    @dispatch(BaseObject)
    def __mul__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('multiply', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __truediv__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __div__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('divide', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __xor__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('xor', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __or__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('or', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __ne__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('compare', self.__class__.__name__, other.__class__.__name__))

    @dispatch(BaseObject)
    def __eq__(self, other):
        raise ValueError('Cannot {} a {} by a {}'.format('compare', self.__class__.__name__, other.__class__.__name__))

    ########################
    # Vectors
    @dispatch(BaseVector4)
    def __add__(self, other):
        return Vector4(super(Vector4, self).__add__(other))

    @dispatch(BaseVector4)
    def __sub__(self, other):
        return Vector4(super(Vector4, self).__sub__(other))

    @dispatch(BaseVector4)
    def __mul__(self, other):
        return Vector4(super(Vector4, self).__mul__(other))

    @dispatch(BaseVector4)
    def __truediv__(self, other):
        return Vector4(super(Vector4, self).__truediv__(other))

    @dispatch(BaseVector4)
    def __div__(self, other):
        return Vector4(super(Vector4, self).__div__(other))

    #@dispatch(BaseVector)
    #def __xor__(self, other):
    #    return self.cross(Vector4(other))

    @dispatch(BaseVector4)
    def __or__(self, other):
        return self.dot(Vector4(other))

    @dispatch(BaseVector4)
    def __ne__(self, other):
        return bool(np.any(super(Vector4, self).__ne__(other)))

    @dispatch(BaseVector4)
    def __eq__(self, other):
        return bool(np.all(super(Vector4, self).__eq__(other)))

    ########################
    # Methods and Properties
    @property
    def negative(self):
        return Vector4(-self)

from .vector3 import Vector3
from .matrix44 import Matrix44



# -*- coding: utf-8 -*-
"""Provide functions for the creation and manipulation of 3D Spheres.

Sphere are represented using a numpy.array of shape (4,).

The first three values are the sphere's position.
The fourth value is the sphere's radius.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from .utils import all_parameters_as_numpy_arrays, parameters_as_numpy_arrays

@parameters_as_numpy_arrays('center')
def create(center=None, radius=1.0, dtype=None):
    if center is None:
        center = [0.,0.,0.]
    return np.array([center[0], center[1], center[2], radius], dtype=dtype)

@parameters_as_numpy_arrays('points')
def create_from_points(points, dtype=None):
    """Creates a sphere centred around 0,0,0 that encompasses
    the furthest point in the provided list.

    :param numpy.array points: An Nd array of vectors.
    :rtype: A sphere as a two value tuple.
    """
    dtype = dtype or points.dtype
    # calculate the lengths of all the points
    # use squared length to save processing
    lengths = np.apply_along_axis(
        np.sum,
        points.ndim - 1,
        points**2
    )

    # find the maximum value
    maximum = lengths.max()

    # square root this, this is the radius
    radius = np.sqrt(maximum)
    return np.array([0.0, 0.0, 0.0, radius], dtype=dtype)

@all_parameters_as_numpy_arrays
def position(sphere):
    """Returns the position of the sphere.

    :param numpy.array sphere: The sphere to extract the position from.
    :rtype: numpy.array
    :return: The centre of the sphere.
    """
    return sphere[:3].copy()

@all_parameters_as_numpy_arrays
def radius(sphere):
    """Returns the radius of the sphere.

    :param numpy.array sphere: The sphere to extract the radius from.
    :rtype: float
    :return: The radius of the sphere.
    """
    return sphere[3]



# -*- coding: utf-8 -*-
"""Provide functions for the creation and manipulation of 2D Rectangles.

Rectangles are represented using a numpy.array of shape (2,2,).

The first value is a vector of x, y position of the rectangle.
The second value is a vector with the width, height of the rectangle.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from .utils import all_parameters_as_numpy_arrays, parameters_as_numpy_arrays


class index:
    #: The index of the position vector within the rectangle
    position = 0

    #: The index of the size vector within the rectangle
    size = 1


def create(x=0., y=0., width=1., height=1., dtype=None):
    """Creates a rectangle from the specified position and sizes.

    This function will interpret the values literally. A negative width
    or height will be represented by the returned value.

    :rtype: numpy.array
    :return: Returns a rectangle with the specified values.
    """
    return np.array([[x, y], [width, height]], dtype=dtype)

def create_zeros(dtype=None):
    return np.zeros((2,2), dtype=dtype)

def create_from_bounds(left, right, bottom, top, dtype=None):
    """Creates a rectangle from the specified boundaries.

    This caters for the left and right, and for the
    top and bottom being swapped.

    :rtype: numpy.array
    :return: Returns a rectangle with the specified values.
        The rectangle will have a positive width and height
        regardless of the values passed in.
    """
    xmin = min(left, right)
    xmax = max(left, right)
    ymin = min(top, bottom)
    ymax = max(top, bottom)

    return create(
        xmin,
        ymin,
        xmax - xmin,
        ymax - ymin,
        dtype
    )

@all_parameters_as_numpy_arrays
def bounds(rect):
    """Returns the absolute boundaries of the rectangle.

    This caters for rectangles with a negative width.

    :rtype: Tuple of 4 floats
    :return: The absolute left, right, bottom and top of the rectangle.
    """
    left = rect[0,0]
    right = rect[0,0] + rect[1,0]
    top = rect[0,1]
    bottom = rect[0,1] + rect[1,1]

    xmin = min(left, right)
    xmax = max(left, right)
    ymin = min(top, bottom)
    ymax = max(top, bottom)

    return xmin, xmax, ymin, ymax

@all_parameters_as_numpy_arrays
def position(rect):
    """Returns the literal position of the rectangle.

    This is the bottom-left point of the rectangle for
    rectangles with positive width and height

    :rtype: numpy.array
    :return: The position of the rectangle.
    """
    return rect[0].copy()

@all_parameters_as_numpy_arrays
def size(rect):
    """Returns the literal size of the rectangle.

    These values may be negative.

    :rtype: numpy.array
    :return: The size of the rectangle.
    """
    return rect[1].copy()

def abs_size(rect):
    """Returns the absolute size of the rectangle.

    :rtype: numpy.array
    :return: The absolute size of the rectangle.
    """
    return np.absolute(rect[1])

def x(rect):
    """Returns the X position of the rectangle.

    This will be the left for rectangles with positive height values.

    :rtype: float
    :return: The X position of the rectangle. This value
        will be further right than the 'right' if the width is negative.
    """
    return rect[0][0]

def y(rect):
    """Returns the Y position of the rectangle.

    This will be the bottom for rectangles with positive height values.

    :rtype: float
    :return: The Y position of the rectangle. This value
        will be above the bottom if the height is negative.
    """
    return rect[0][1]

def width(rect):
    """Returns the literal width of the rectangle.

    :rtype: float
    :return: The width of the rectangle. This can be a
        negative value.
    """
    return rect[1][0]

def abs_width(rect):
    """Returns the absolute width of the rectangle.

    This caters for rectangles with a negative width.

    :rtype: float
    :return: The absolute width of the rectangle.
    """
    return abs(width(rect))

def height(rect):
    """Returns the literal height of the rectangle.

    :rtype: float
    :return: The height of the rectangle. This can be a
        negative value.
    """
    return rect[1][1]

def abs_height(rect):
    """Returns the absolute height of the rectangle.

    This caters for rectangles with a negative height.

    :rtype: float
    :return: The absolute height of the rectangle.
    """
    return abs(height(rect))

def top(rect):
    """Returns the top most Y value of the rectangle.

    This caters for rectangles with a negative height.

    :rtype: float
    :return: The biggest Y value.
    """
    return max(
        rect[0][1],
        rect[0][1] + rect[1][1]
       )

def bottom(rect):
    """Returns the bottom most Y value of the rectangle.

    This caters for rectangles with a negative height.

    :rtype: float
    :return: The smallest Y value.
    """
    return min(
        rect[0][1],
        rect[0][1] + rect[1][1]
       )

def left(rect):
    """Returns the left most X value of the rectangle.

    This caters for rectangles with a negative width.

    :rtype: float
    :return: The smallest X value.
    """
    return min(
        rect[0][0],
        rect[0][0] + rect[1][0]
       )

def right(rect):
    """Returns the right most X value of the rectangle.

    This caters for rectangles with a negative width.

    :rtype: float
    :return: The biggest X value.
    """
    return max(
        rect[0][0],
        rect[0][0] + rect[1][0]
       )

@parameters_as_numpy_arrays('rect')
def scale_by_vector(rect, vec):
    """Scales a rectangle by a 2D vector.

    Note that this will also scale the X,Y
    value of the rectangle, which will cause
    the rectangle to move, not just increase
    in size.

    :param numpy.array rect: the rectangle to scale.
        Both x,y and width,height will be scaled.
    :param vec: A 2D vector to scale the rect by.
    :rtype: numpy.array.
    """
    return rect * vec[:2]

def aspect_ratio(rect):
    width = float(abs_width(rect))
    height = float(abs_height(rect))
    return width / height


